"""
This module implements the Room class.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from . import app as _app
    from . import thermostat

import datetime

from .. import common
from . import expr, schedule, util, window_sensor


class Room:
    """A room to be controlled by Heaty."""

    # pylint: disable=too-many-instance-attributes

    def __init__(self, name: str, cfg: dict, app: "_app.HeatyApp") -> None:
        self.name = name
        self.cfg = cfg
        self.app = app
        self.thermostats = []  # type: T.List[thermostat.Thermostat]
        self.window_sensors = []  # type: T.List[window_sensor.WindowSensor]
        self.schedule = None  # type: T.Optional[schedule.Schedule]

        self.wanted_temp = None  # type: T.Optional[expr.Temp]
        self.reschedule_timer = None  # type: T.Optional[uuid.UUID]
        self.current_schedule_temp = None  # type: T.Optional[expr.Temp]
        self.current_schedule_rule = None  # type: T.Optional[schedule.Rule]

    def __repr__(self) -> str:
        return "<Room {}>".format(str(self))

    def __str__(self) -> str:
        return self.cfg.get("friendly_name", self.name)

    @util.modifies_state
    def _reschedule_timer_cb(self, kwargs: dict) -> None:
        """Is called whenever a re-schedule timer fires."""

        self.log("Re-schedule timer fired.",
                 level="DEBUG")

        self.reschedule_timer = None

        # invalidate cached temp/rule
        self.current_schedule_temp = None
        self.current_schedule_rule = None

        self.set_scheduled_temp()

    def _schedule_timer_cb(self, kwargs: dict) -> None:
        """Is called whenever a schedule timer fires."""

        self.log("Schedule timer fired.",
                 level="DEBUG")
        self.set_scheduled_temp()

    def initialize(self) -> None:
        """Should be called after all schedules, thermostats and window
        sensors have been added in order to register state listeners
        and timers."""

        for therm in self.thermostats:
            therm.initialize()

        for wsensor in self.window_sensors:
            wsensor.initialize()

        if self.schedule:
            self.log("Creating schedule timers.", level="DEBUG")
            # We collect the times in a set first to avoid registering
            # multiple timers for the same time.
            times = set()  # type: T.Set[datetime.time]
            for rule in self.schedule.unfold():
                times.update((rule.start_time, rule.end_time))

            # Now register a timer for each time a rule starts or ends.
            for _time in times:
                self.log("Registering timer at {}.".format(_time),
                         level="DEBUG")
                self.app.run_daily(self._schedule_timer_cb, _time)

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the room to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.app.log(msg, *args, **kwargs)

    @util.modifies_state
    def set_temp(
            self, target_temp: expr.Temp, scheduled: bool = False,
            force_resend: bool = False
    ) -> None:
        """Sets the given target temperature for all thermostats in the
        room. If scheduled is True, a disabled master switch prevents
        setting the temperature.
        Temperatures won't be send to thermostats redundantly unless
        force_resend is True."""

        if scheduled and \
           not self.app.master_switch_enabled():
            return

        self.wanted_temp = target_temp

        changed = False
        for therm in self.thermostats:
            result = therm.set_temp(target_temp, force_resend=force_resend)
            changed = changed or bool(result)

        if changed:
            self.log("Temperature set to {}.  <{}>"
                     .format(target_temp,
                             "scheduled" if scheduled else "manual"),
                     prefix=common.LOG_PREFIX_OUTGOING)

    def eval_temp_expr(
            self, temp_expr: expr.EXPR_TYPE
    ) -> T.Optional[expr.ResultBase]:
        """This is a wrapper around expr.eval_temp_expr that adds the
        app object, the room name  and some helpers to the evaluation
        environment, as well as all configured
        temp_expression_modules. It also catches and logs any
        exception which is raised during evaluation. In this case,
        None is returned."""

        # use date/time provided by appdaemon to support time-traveling
        now = self.app.datetime()
        extra_env = {
            "app": self.app,
            "schedule_snippets": self.app.cfg["schedule_snippets"],
            "room_name": self.name,
            "now": now,
            "date": now.date(),
            "time": now.time(),
        }
        extra_env.update(self.app.temp_expression_modules)

        try:
            return expr.eval_temp_expr(temp_expr, extra_env=extra_env)
        except Exception as err:  # pylint: disable=broad-except
            self.log("Error while evaluating temperature expression: "
                     "{}".format(repr(err)),
                     level="ERROR")
            return None

    def eval_schedule(
            self, sched: schedule.Schedule, when: datetime.datetime
    ) -> T.Optional[T.Tuple[expr.Temp, schedule.Rule]]:
        """Evaluates a schedule, computing the temperature for the time
        the given datetime object represents. The temperature and the
        matching rule are returned.
        If no temperature could be found in the schedule (e.g. all
        rules evaluate to Ignore()), None is returned."""

        result_sum = expr.Add(0)
        rules = list(sched.matching_rules(when))
        idx = 0
        while idx < len(rules):
            rule = rules[idx]
            idx += 1

            result = self.eval_temp_expr(rule.temp_expr)
            self.log("Evaluated temperature expression {} to {}."
                     .format(repr(rule.temp_expr_raw), result),
                     level="DEBUG")

            if result is None:
                self.log("Skipping rule with faulty temperature "
                         "expression: {}"
                         .format(rule.temp_expr_raw))
                continue

            if isinstance(result, expr.Break):
                # abort, don't change temperature
                self.log("Aborting scheduling due to Break().",
                         level="DEBUG")
                return None

            if isinstance(result, expr.Ignore):
                self.log("Skipping this rule.",
                         level="DEBUG")
                continue

            if isinstance(result, expr.IncludeSchedule):
                self.log("Inserting sub-schedule.",
                         level="DEBUG")
                _rules = result.schedule.matching_rules(self.app.datetime())
                for _idx, _rule in enumerate(_rules):
                    rules.insert(idx + _idx, _rule)
                continue

            if isinstance(result, expr.AddibleMixin):
                result_sum += result

            if isinstance(result_sum, expr.Result):
                return result_sum.temp, rule

        return None

    def get_scheduled_temp(
            self,
    ) -> T.Optional[T.Tuple[expr.Temp, schedule.Rule]]:
        """Computes and returns the temperature that is configured for
        the current date and time. The second return value is the rule
        which generated the result.
        If no temperature could be found in the schedule (e.g. all
        rules evaluate to Ignore()), None is returned."""

        if self.schedule is None:
            return None
        return self.eval_schedule(self.schedule, self.app.datetime())

    def set_scheduled_temp(self, force_resend: bool = False) -> None:
        """Sets the temperature that is configured for the current
        date and time. If the master switch is turned off, this won't
        do anything.
        This method won't re-schedule if a re-schedule timer runs.
        It will also detect when neither the rule nor the result
        of its temperature expression changed compared to the last run
        and prevent re-setting the temperature in that case.
        If force_resend is True, and the temperature didn't
        change, it is sent to the thermostats anyway.
        In case of an open window, temperature is cached and not sent."""

        if not self.app.master_switch_enabled():
            return

        if self.reschedule_timer:
            # don't schedule now, wait for the timer instead
            self.log("Not scheduling now due to a running re-schedule "
                     "timer.",
                     level="DEBUG")
            return

        result = self.get_scheduled_temp()
        if result is None:
            self.log("No suitable temperature found in schedule.",
                     level="DEBUG")
            return

        temp, rule = result
        if temp == self.current_schedule_temp and \
           rule is self.current_schedule_rule and \
           not force_resend:
            # temp and rule didn't change, what means that the
            # re-scheduling wasn't necessary and was e.g. caused
            # by a daily timer which doesn't count for today
            return

        self.current_schedule_temp = temp
        self.current_schedule_rule = rule

        if self.get_open_windows():
            self.log("Caching and not setting temperature due to an "
                     "open window.")
            self.wanted_temp = temp
        else:
            self.set_temp(temp, scheduled=True, force_resend=force_resend)

    def set_manual_temp(
            self, temp_expr: expr.EXPR_TYPE, force_resend: bool = False,
            reschedule_delay: T.Union[float, int, None] = None
    ) -> None:
        """Evaluates the given temperature expression and sets the result.
        If the master switch is turned off, this won't do anything.
        If force_resend is True, and the temperature didn't
        change, it is sent to the thermostats anyway.
        An existing re-schedule timer is cancelled and a new one is
        started if re-schedule timers are configured. reschedule_delay,
        if given, overwrites the value configured for the room.
        In case of an open window, temperature is cached and not sent."""

        if not self.app.master_switch_enabled():
            return

        result = self.eval_temp_expr(temp_expr)
        self.log("Evaluated temperature expression {} to {}."
                 .format(repr(temp_expr), repr(result)),
                 level="DEBUG")

        temp = None
        if isinstance(result, expr.IncludeSchedule):
            _result = self.eval_schedule(result.schedule, self.app.datetime())
            if _result is not None:
                temp = _result[0]
        elif isinstance(result, expr.Result):
            temp = result.temp

        if temp is None:
            self.log("Ignoring temperature expression.")
            return

        if self.get_open_windows():
            self.log("Caching and not setting temperature due to an"
                     "open window.")
            self.wanted_temp = temp
        else:
            self.set_temp(temp, scheduled=False, force_resend=force_resend)

        self.update_reschedule_timer(reschedule_delay=reschedule_delay)

    @util.modifies_state
    def cancel_reschedule_timer(self) -> bool:
        """Cancels the reschedule timer for this room, if one
        exists. Returns whether a timer has been cancelled."""

        timer = self.reschedule_timer
        if timer is None:
            return False

        self.app.cancel_timer(timer)
        self.reschedule_timer = None
        self.log("Cancelled re-schedule timer.", level="DEBUG")
        return True

    def get_open_windows(self) -> T.List[window_sensor.WindowSensor]:
        """Returns a list of window sensors in this room which
        currently report to be open,"""

        _open = filter(lambda sensor: sensor.is_open(),
                       self.window_sensors)
        return list(_open)

    def check_for_open_window(self) -> bool:
        """Checks whether a window is open in this room and,
        if so, turns the heating off there. The value stored in
        self.wanted_temp is restored after the heating
        has been turned off. It returns True if a window is open,
        False otherwise."""

        if self.get_open_windows():
            # window is open, turn heating off
            orig_temp = self.wanted_temp
            off_temp = self.app.cfg["off_temp"]
            if orig_temp != off_temp:
                self.log("Turning heating off due to an open "
                         "window.",
                         prefix=common.LOG_PREFIX_OUTGOING)
                self.set_temp(off_temp, scheduled=False)
                self.wanted_temp = orig_temp
            return True
        return False

    def notify_set_temp_event(
            self, temp_expr: expr.EXPR_TYPE, force_resend: bool = False,
            reschedule_delay: T.Union[float, int, None] = None
    ) -> None:
        """Handles a heaty_set_temp event for this room."""

        self.log("heaty_set_temp event received, temperature: {}"
                 .format(repr(temp_expr)))
        self.set_manual_temp(temp_expr, force_resend=force_resend,
                             reschedule_delay=reschedule_delay)

    def notify_temp_changed(
            self, temp: expr.Temp, no_reschedule: bool = False
    ) -> None:
        """Should be called when the temperature has been changed
        externally, e.g. by manual adjustment at a thermostat.
        Setting no_reschedule prevents re-scheduling."""

        if self.get_open_windows():
            # After window has been opened and heating turned off,
            # thermostats usually report to be off, but we don't
            # care to not mess up self.wanted_temp and prevent
            # replication.
            return

        self.wanted_temp = temp

        if not self.app.master_switch_enabled():
            return

        if self.cfg["replicate_changes"] and len(self.thermostats) > 1 and \
           self.app.master_switch_enabled():
            self.log("Propagating the change to all thermostats "
                     "in the room.",
                     prefix=common.LOG_PREFIX_OUTGOING)
            self.set_temp(temp, scheduled=False)

        if not no_reschedule:
            self.update_reschedule_timer()

    @util.modifies_state
    def update_reschedule_timer(
            self, reschedule_delay: T.Union[float, int, None] = None,
            force: bool = False,
    ) -> None:
        """This method cancels an existing re-schedule timer first.
        Then, it checks if either force is set or the wanted
        temperature differs from the scheduled temperature. If
        so, a new timer is created according to the room's
        settings. reschedule_delay, if given, overwrites the value
        configured for the room."""

        self.cancel_reschedule_timer()

        if not self.app.master_switch_enabled():
            return

        if reschedule_delay is None:
            reschedule_delay = self.cfg["reschedule_delay"]

        wanted = self.wanted_temp
        result = self.get_scheduled_temp()
        if not reschedule_delay or \
           (not force and result and wanted == result[0]):
            return

        delta = datetime.timedelta(minutes=reschedule_delay)
        when = self.app.datetime() + delta
        self.log("Re-scheduling not before {} ({})."
                 .format(util.format_time(when.time()), delta))
        timer = self.app.run_at(self._reschedule_timer_cb, when)
        self.reschedule_timer = timer



"""
This module implements the WindowSensor class.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from . import room as _room

from .. import common
from . import util


class WindowSensor:
    """A sensor for Heaty's open window detection."""

    def __init__(self, entity_id: str, cfg: dict, room: "_room.Room") -> None:
        self.entity_id = entity_id
        self.cfg = cfg
        self.room = room
        self.app = room.app

    def __repr__(self) -> str:
        return "<WindowSensor {}, {}>".format(
            str(self), "open" if self.is_open() else "closed"
        )

    def __str__(self) -> str:
        return self.cfg.get("friendly_name", self.entity_id)

    @util.modifies_state
    def _state_cb(
            self, entity: str, attr: str,
            old: T.Optional[dict], new: T.Optional[dict],
            kwargs: dict
    ) -> None:
        """Is called when the window sensor's state has changed.
        This method handles the window open/closed detection and
        performs actions accordingly."""

        action = "opened" if self.is_open() else "closed"
        self.log("State is now {}.".format(new),
                 level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)
        self.room.log("Window has been {}.".format(action),
                      prefix=common.LOG_PREFIX_INCOMING)

        if not self.app.master_switch_enabled():
            self.log("Master switch is off, ignoring window event.")
            return

        if action == "opened":
            # turn heating off, but store the original temperature
            self.room.check_for_open_window()
        elif not self.room.get_open_windows():
            # all windows closed
            # restore temperature from before opening the window
            orig_temp = self.room.wanted_temp
            # could be None if we didn't know the temperature before
            # opening the window
            if orig_temp is None:
                self.log("Restoring temperature from schedule.",
                         level="DEBUG")
                self.room.set_scheduled_temp()
            else:
                self.log("Restoring temperature to {}.".format(orig_temp),
                         level="DEBUG")
                self.room.set_temp(orig_temp, scheduled=False)

    def initialize(self) -> None:
        """Should be called in order to register state listeners and
        timers."""

        self.log("Registering window sensor state listener, delay = {}."
                 .format(self.cfg["delay"]),
                 level="DEBUG")
        self.app.listen_state(self._state_cb, self.entity_id,
                              duration=self.cfg["delay"])

    def is_open(self) -> bool:
        """Returns whether the sensor reports open or not."""

        open_state = self.cfg["open_state"]
        states = []
        if isinstance(open_state, list):
            states.extend(open_state)
        else:
            states.append(open_state)
        return self.app.get_state(self.entity_id) in states

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the window sensor to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.room.log(msg, *args, **kwargs)



"""
This module implements the Thermostat class.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    import uuid
    from . import room as _room

from .. import common
from . import expr, util


class Thermostat:
    """A thermostat to be controlled by Heaty."""

    def __init__(self, entity_id: str, cfg: dict, room: "_room.Room") -> None:
        self.entity_id = entity_id
        self.cfg = cfg
        self.room = room
        self.app = room.app
        self.current_temp = None  # type: T.Optional[expr.Temp]
        self.wanted_temp = None  # type: T.Optional[expr.Temp]
        self.resend_timer = None  # type: T.Optional[uuid.UUID]

    def __repr__(self) -> str:
        return "<Thermostat {}>".format(str(self))

    def __str__(self) -> str:
        return self.cfg.get("friendly_name", self.entity_id)

    def _check_config_plausibility(self) -> None:
        """Is called during initialization to warn the user about some
        possible common configuration mistakes."""

        state = self.app.get_state(self.entity_id, attribute="all")
        if not state:
            self.log("Thermostat couldn't be found.", level="WARNING")
            return

        state = (state or {}).get("attributes", {})

        required_attrs = [self.cfg["temp_state_attr"]]
        if self.cfg["supports_opmodes"]:
            required_attrs.append(self.cfg["opmode_state_attr"])
        for attr in required_attrs:
            if attr not in state:
                self.log("Thermostat has no attribute named {}. "
                         "Available attributes are {}. "
                         "Please check your config!"
                         .format(repr(attr), list(state.keys())),
                         level="WARNING")

        allowed_opmodes = state.get("operation_list")
        if not self.cfg["supports_opmodes"]:
            if allowed_opmodes:
                self.log("Operation mode support has been disabled, "
                         "but the following modes seem to be supported: {} "
                         "Maybe disabling it was a mistake?"
                         .format(allowed_opmodes),
                         level="WARNING")
            return
        if not allowed_opmodes:
            self.log("Attributes for thermostat contain no "
                     "'operation_list', Consider disabling "
                     "operation mode support.",
                     level="WARNING")
            return
        for opmode in (self.cfg["opmode_heat"], self.cfg["opmode_off"]):
            if opmode not in allowed_opmodes:
                self.log("Thermostat doesn't seem to support the "
                         "operation mode {}, supported modes are: {}. "
                         "Please check your config!"
                         .format(opmode, allowed_opmodes),
                         level="WARNING")

    @util.modifies_state
    def _state_cb(
            self, entity: str, attr: str,
            old: T.Optional[dict], new: T.Optional[dict],
            kwargs: dict, no_reschedule: bool = False
    ) -> None:
        """Is called when the thermostat's state changes.
        This method fetches the set target temperature from the
        thermostat and reacts accordingly."""

        attrs = (new or {}).get("attributes", {})

        opmode = attrs.get(self.cfg["opmode_state_attr"])
        self.log("Attribute {} is {}."
                 .format(self.cfg["opmode_state_attr"], opmode),
                 level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)

        _temp = None
        if self.cfg["supports_opmodes"]:
            if opmode is None:
                # don't consider this thermostat
                return
            elif opmode == self.cfg["opmode_off"]:
                _temp = expr.Off()

        if _temp is None:
            _temp = attrs.get(self.cfg["temp_state_attr"])
            self.log("Attribute {} is {}."
                     .format(self.cfg["temp_state_attr"], _temp),
                     level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)

        try:
            temp = expr.Temp(_temp)
        except ValueError:
            # not a valid temperature, don't consider this thermostat
            return

        if temp == self.wanted_temp:
            # thermostat adapted to the temperature we want,
            # cancel any re-send timer
            self.cancel_resend_timer()

        if temp == self.current_temp:
            # nothing changed, hence no further actions needed
            return

        self.current_temp = temp

        self.log("Received target temperature {}."
                 .format(repr(temp)),
                 prefix=common.LOG_PREFIX_INCOMING)

        temp -= self.cfg["delta"]

        self.room.notify_temp_changed(temp, no_reschedule=no_reschedule)

    def initialize(self) -> None:
        """Should be called in order to register state listeners and
        timers."""

        self._check_config_plausibility()

        # only consider one thermostat per room
        if self.room.wanted_temp is None:
            self.log("Getting current temperature from thermostat.",
                     level="DEBUG")
            state = self.app.get_state(self.entity_id, attribute="all")
            if state is None:
                # unknown entity
                self.log("State for thermostat is None, ignoring it.",
                         level="WARNING")
            else:
                # populate self.current_temp by simulating a state change
                self._state_cb(self.entity_id, "all", state, state, {},
                               no_reschedule=True)

        self.log("Registering thermostat state listener.", level="DEBUG")
        self.app.listen_state(self._state_cb, self.entity_id, attribute="all")

    def log(self, msg: str, *args: T.Any, **kwargs: T.Any) -> None:
        """Prefixes the thermostat to log messages."""
        msg = "[{}] {}".format(self, msg)
        self.room.log(msg, *args, **kwargs)

    def cancel_resend_timer(self) -> bool:
        """Cancel the resend timer for this thermostat, if one exists.
        Returns whether a timer was cancelled."""

        timer = self.resend_timer
        if timer is None:
            return False

        self.app.cancel_timer(timer)
        self.resend_timer = None
        self.log("Cancelled resend timer.", level="DEBUG")
        return True

    def is_synced(self) -> bool:
        """Returns whether the thermostat's target temperature is the
        wanted temperature and no re-sending is in progress."""

        return self.resend_timer is None and \
               self.current_temp is not None and \
               self.current_temp == self.wanted_temp

    @util.modifies_state
    def set_temp(
            self, target_temp: expr.Temp, force_resend: bool = False
    ) -> T.Optional[expr.Temp]:
        """Sets the given target temperature on this thermostat
        The temperature should be the desired room temperature,
        without delta added. This method will try best to achieve
        the closest possible temperature supported by this particular
        thermostat.
        A temperature won't be send to the thermostat redundantly
        unless force_resend is True.
        The return value is either the actually set temperature or
        None, if nothing has been sent."""

        if target_temp.is_off():
            temp = None
            opmode = self.cfg["opmode_off"]
        else:
            temp = target_temp + self.cfg["delta"]
            if isinstance(self.cfg["min_temp"], expr.Temp) and \
               temp < self.cfg["min_temp"]:
                temp = None
                opmode = self.cfg["opmode_off"]
            else:
                opmode = self.cfg["opmode_heat"]

        if not self.cfg["supports_opmodes"]:
            if opmode == self.cfg["opmode_off"]:
                self.log("Not turning off because it doesn't support "
                         "operation modes.",
                         level="DEBUG")
                if self.cfg["min_temp"] is not None:
                    self.log("Setting to minimum supported temperature "
                             "instead.",
                             level="DEBUG")
                    temp = self.cfg["min_temp"]

            opmode = None

        if opmode is None and temp is None:
            self.log("Nothing to send to thermostat.",
                     level="DEBUG")
            return None

        self.wanted_temp = temp or expr.Temp(expr.OFF)
        if not force_resend and self.is_synced():
            self.log("Not sending temperature redundantly.",
                     level="DEBUG")
            return None

        left_retries = self.cfg["set_temp_retries"]
        self.cancel_resend_timer()
        timer = self.app.run_in(
            self._set_temp_resend_cb, 1,
            left_retries=left_retries,
            opmode=opmode, temp=temp
        )
        self.resend_timer = timer

        return self.wanted_temp

    def _set_temp_resend_cb(self, kwargs: dict) -> None:
        """This callback sends the operation_mode and temperature to the
        thermostat. Expected values for kwargs are:
        - opmode and temp (incl. delta)
        - left_retries (after this round)"""

        opmode = kwargs["opmode"]
        temp = kwargs["temp"]
        left_retries = kwargs["left_retries"]

        self.cancel_resend_timer()

        self.log("Setting {} = {}, {} = {}, left retries = {}"
                 .format(self.cfg["temp_service_attr"],
                         temp if temp is not None else "<unset>",
                         self.cfg["opmode_service_attr"],
                         opmode if opmode is not None else "<unset>",
                         left_retries),
                 level="DEBUG", prefix=common.LOG_PREFIX_OUTGOING)

        if opmode is not None:
            attrs = {"entity_id": self.entity_id,
                     self.cfg["opmode_service_attr"]: opmode}
            self.app.call_service(self.cfg["opmode_service"], **attrs)
        if temp is not None:
            attrs = {"entity_id": self.entity_id,
                     self.cfg["temp_service_attr"]: temp.value}
            self.app.call_service(self.cfg["temp_service"], **attrs)

        if not left_retries:
            return

        interval = self.cfg["set_temp_retry_interval"]
        self.log("Re-sending in {} seconds."
                 .format(interval),
                 level="DEBUG")
        timer = self.app.run_in(
            self._set_temp_resend_cb, interval,
            left_retries=left_retries - 1,
            opmode=opmode, temp=temp)
        self.resend_timer = timer



"""
Module containing functionality to evaluate temperature expressions.
"""

import typing as T
if T.TYPE_CHECKING:
    # pylint: disable=cyclic-import,unused-import
    from . import schedule
import types

import datetime
import functools


__all__ = ["Add", "Break", "Ignore", "IncludeSchedule", "Off", "OFF",
           "Result", "Temp"]


# type of an evaluable expression
EXPR_TYPE = T.Union[str, types.CodeType, "Temp"]


class AddibleMixin:
    """Mixin that makes a temperature expression result addible."""
    pass

class ResultBase:
    """Holds the result of a temperature expression."""

    def __init__(self, temp: T.Any) -> None:
        self.temp = Temp(temp)  # type: T.Optional[Temp]

    def __eq__(self, other: T.Any) -> bool:
        return type(self) is type(other) and self.temp == other.temp


class Result(ResultBase, AddibleMixin):
    """Final result of a temperature expression."""

    def __repr__(self) -> str:
        return "{}".format(self.temp)

class Add(ResultBase, AddibleMixin):
    """Result of a temperature expression that is intended to be added
    to the result of a consequent expression."""

    def __add__(self, other: ResultBase) -> ResultBase:
        if not isinstance(other, AddibleMixin):
            raise TypeError("can't add {} and {}"
                            .format(repr(type(self)), repr(type(other))))

        return type(other)(self.temp + other.temp)

    def __repr__(self) -> str:
        return "Add({})".format(self.temp)

class Break(ResultBase):
    """Result of a temperature expression that should abort scheduling and
    leave the temperature unchanged."""

    def __init__(self) -> None:
        # pylint: disable=super-init-not-called
        self.temp = None

    def __repr__(self) -> str:
        return "Break()"

class Ignore(ResultBase):
    """Result of a temperature expression which should be ignored."""

    def __init__(self) -> None:
        # pylint: disable=super-init-not-called
        self.temp = None

    def __repr__(self) -> str:
        return "Ignore()"

class IncludeSchedule(ResultBase):
    """Result that includes a schedule for processing."""

    def __init__(self, sched: "schedule.Schedule") -> None:
        # pylint: disable=super-init-not-called
        self.schedule = sched

    def __repr__(self) -> str:
        return "IncludeSchedule({})".format(self.schedule)


class Off:
    """A special value Temp() may be initialized with in order to turn
    a thermostat off."""

    def __add__(self, other: T.Any) -> "Off":
        return self

    def __eq__(self, other: T.Any) -> bool:
        return isinstance(other, Off)

    def __neg__(self) -> "Off":
        return self

    def __repr__(self) -> str:
        return "OFF"

    def __sub__(self, other: T.Any) -> "Off":
        return self

OFF = Off()


@functools.total_ordering
class Temp:
    """A class holding a temperature value."""

    def __init__(self, temp_value: T.Any) -> None:
        if isinstance(temp_value, Temp):
            # Just copy the value over.
            parsed = self.parse_temp(temp_value.value)
        else:
            parsed = self.parse_temp(temp_value)

        if parsed is None:
            raise ValueError("{} is no valid temperature"
                             .format(repr(temp_value)))

        self.value = parsed  # type: T.Union[float, Off]

    def __add__(self, other: T.Any) -> "Temp":
        if isinstance(other, (float, int)):
            other = Temp(other)
        elif not isinstance(other, Temp):
            raise TypeError("can't add {} and {}"
                            .format(repr(type(self)), repr(type(other))))

        # OFF + something is OFF
        if self.is_off() or other.is_off():
            return Temp(Off())

        return Temp(self.value + other.value)

    def __eq__(self, other: T.Any) -> bool:
        return isinstance(other, Temp) and self.value == other.value

    def __lt__(self, other: T.Any) -> bool:
        if isinstance(other, (float, int)):
            other = Temp(other)

        if type(self) is not type(other):
            raise TypeError("can't compare {} and {}"
                            .format(repr(type(self)), repr(type(other))))

        if not self.is_off() and other.is_off():
            return False
        if self.is_off() and not other.is_off() or \
           self.value < other.value:
            return True
        return False

    def __neg__(self) -> "Temp":
        return Temp(-self.value)  # pylint: disable=invalid-unary-operand-type

    def __repr__(self) -> str:
        return repr(self.value)

    def __str__(self) -> str:
        return str(self.value)

    def __sub__(self, other: T.Any) -> "Temp":
        return self.__add__(-other)

    def is_off(self) -> bool:
        """Returns whether this temperature means OFF."""
        return isinstance(self.value, Off)

    @staticmethod
    def parse_temp(value: T.Any) -> T.Union[float, Off, None]:
        """Converts the given value to a valid temperature of type float
        or Off().
        If value is a string, all whitespace is removed first.
        If conversion is not possible, None is returned."""

        if isinstance(value, str):
            value = "".join(value.split())
            if value.upper() == "OFF":
                return Off()

        if isinstance(value, Off):
            return Off()

        try:
            return float(value)
        except (ValueError, TypeError):
            return None


def build_time_expression_env() -> T.Dict[str, T.Any]:
    """This function builds and returns an environment usable as globals
    for the evaluation of a time expression. It will add all members
    of this module's __all__ to the environment."""

    env = {"datetime": datetime}
    for name in __all__:
        env[name] = globals()[name]
    return env

def eval_temp_expr(
        temp_expr: EXPR_TYPE,
        extra_env: T.Optional[T.Dict[str, T.Any]] = None
) -> ResultBase:
    """This method evaluates the given temperature expression.
    The evaluation result is returned. The items of the extra_env
    dict are added to the globals available during evaluation.
    If the expression is a Temp object already, it's just packed into
    a Result and returned directly."""

    # pylint: disable=eval-used

    if isinstance(temp_expr, Temp):
        return Result(temp_expr)

    env = build_time_expression_env()
    if extra_env:
        env.update(extra_env)

    eval_result = eval(temp_expr, env)
    if isinstance(eval_result, ResultBase):
        result = eval_result
    else:
        result = Result(eval_result)
    return result



"""
This module contains the CONFIG_SCHEMA for validation with voluptuous.
"""

import typing as T

import voluptuous as vol

from . import expr, schedule, thermostat, util, window_sensor
from . import room as _room


def build_schedule_rule(rule: dict) -> schedule.Rule:
    """Builds and returns a schedule rule from the given rule
    definition."""

    constraints = {}
    for name, value in rule.items():
        if name in schedule.Rule.CONSTRAINTS:
            constraints[name] = value

    start_time = rule.get("start")
    if start_time is not None:
        start_time = util.parse_time_string(start_time)

    end_time = rule.get("end")
    if end_time is not None:
        end_time = util.parse_time_string(end_time)
    end_plus_days = rule["end_plus_days"]

    temp_expr = rule["temp"]

    return schedule.Rule(temp_expr=temp_expr,
                         start_time=start_time,
                         end_time=end_time,
                         end_plus_days=end_plus_days,
                         constraints=constraints)

def build_schedule(rules: T.Iterable[dict]) -> schedule.Schedule:
    """Compiles the given rules and returns a schedule containing them."""

    sched = schedule.Schedule()
    for rule in rules:
        sched.items.append(build_schedule_rule(rule))
    return sched

def config_post_hook(cfg: dict) -> dict:
    """Creates room and thermostat objects after config has been parsed."""

    # Compile the pre/post schedules.
    cfg["schedule_prepend"] = build_schedule(cfg["schedule_prepend"])
    cfg["schedule_append"] = build_schedule(cfg["schedule_append"])

    # Build room objects.
    rooms = []
    for room_name, room_data in cfg["rooms"].items():
        therms = {}
        wsensors = {}

        # copy settings from defaults sections to this room
        for therm_name, therm_data in room_data["thermostats"].items():
            for key, val in cfg["thermostat_defaults"].items():
                therm_data.setdefault(key, val)
            therm_data = THERMOSTAT_SCHEMA(therm_data)
            therms[therm_name] = therm_data
        for wsensor_name, wsensor_data in room_data["window_sensors"].items():
            for key, val in cfg["window_sensor_defaults"].items():
                wsensor_data.setdefault(key, val)
            wsensor_data = WINDOW_SENSOR_SCHEMA(wsensor_data)
            wsensors[wsensor_name] = wsensor_data

        # Compile the room's schedule.
        sched = build_schedule(room_data["schedule"])
        sched.items.insert(0, cfg["schedule_prepend"])
        sched.items.append(cfg["schedule_append"])

        del room_data["thermostats"]
        del room_data["window_sensors"]
        del room_data["schedule"]

        room = _room.Room(room_name, room_data, cfg["_app"])
        rooms.append(room)

        # Create thermostat and window sensor objects and attach to room.
        for therm_name, therm_data in therms.items():
            therm = thermostat.Thermostat(therm_name, therm_data, room)
            room.thermostats.append(therm)
        for wsensor_name, wsensor_data in wsensors.items():
            wsensor = window_sensor.WindowSensor(
                wsensor_name, wsensor_data, room
            )
            room.window_sensors.append(wsensor)

        room.schedule = sched

    del cfg["rooms"], cfg["schedule_prepend"], cfg["schedule_append"]
    cfg["_app"].rooms = rooms

    return cfg


ENTITY_ID_SCHEMA = vol.Schema(vol.Match(r"^[A-Za-z_]+\.[A-Za-z0-9_]+$"))
PYTHON_VAR_SCHEMA = vol.Schema(vol.Match(r"^[a-zA-Z_]+[a-zA-Z0-9_]*$"))
RANGE_STRING_SCHEMA = vol.Schema(vol.All(
    vol.Any(
        int,
        vol.Match(r"^ *\d+( *\- *\d+)?( *\, *\d+( *\- *\d+)?)* *$"),
    ),
    util.expand_range_string,
))
PARTIAL_DATE_SCHEMA = vol.Schema({
    vol.Optional("year"): vol.All(int, vol.Range(min=1970, max=9999)),
    vol.Optional("month"): vol.All(int, vol.Range(min=1, max=12)),
    vol.Optional("day"): vol.All(int, vol.Range(min=1, max=31)),
})
TIME_SCHEMA = vol.Schema(vol.Match(r"^ *([01]\d|2[0123]) *\: *([012345]\d) *$"))
TEMP_SCHEMA = vol.Schema(vol.All(
    vol.Any(float, int, expr.Off, vol.All(str, lambda v: v.upper(), "OFF")),
    lambda v: expr.Temp(v),  # pylint: disable=unnecessary-lambda
))
TEMP_EXPRESSION_SCHEMA = vol.Schema(str)

# This schema does no real validation and default value insertion,
# it just ensures a dictionary containing dictionaries is returned.
DICTS_IN_DICT_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {vol.Extra: lambda v: v or {}},
))

TEMP_EXPRESSION_MODULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "as": PYTHON_VAR_SCHEMA,
    },
))
TEMP_EXPRESSION_MODULES_SCHEMA = vol.Schema({
    vol.Extra: lambda v: TEMP_EXPRESSION_MODULE_SCHEMA(v or {}),
})

THERMOSTAT_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("delta", default=0): vol.Any(float, int),
        vol.Optional("min_temp", default=None): vol.Any(
            vol.All(TEMP_SCHEMA, vol.NotIn([expr.Temp(expr.OFF)])),
            None,
        ),
        vol.Optional("set_temp_retries", default=10):
            vol.All(int, vol.Range(min=-1)),
        vol.Optional("set_temp_retry_interval", default=30):
            vol.All(int, vol.Range(min=1)),
        vol.Optional("supports_opmodes", default=True): bool,
        vol.Optional("opmode_heat", default="Heat"): str,
        vol.Optional("opmode_off", default="Off"): str,
        vol.Optional("opmode_service", default="climate/set_operation_mode"):
            str,
        vol.Optional("opmode_service_attr", default="operation_mode"): str,
        vol.Optional("opmode_state_attr", default="operation_mode"): str,
        vol.Optional("temp_service", default="climate/set_temperature"): str,
        vol.Optional("temp_service_attr", default="temperature"): str,
        vol.Optional("temp_state_attr", default="temperature"): str,
    },
))

STATE_SCHEMA = vol.Schema(vol.Any(float, int, str))
WINDOW_SENSOR_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("delay", default=10): vol.All(int, vol.Range(min=0)),
        vol.Optional("open_state", default="on"):
            vol.Any(STATE_SCHEMA, [STATE_SCHEMA]),
    },
))

SCHEDULE_RULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "temp": vol.Any(TEMP_SCHEMA, TEMP_EXPRESSION_SCHEMA),
        vol.Optional("start", default=None): vol.Any(TIME_SCHEMA, None),
        vol.Optional("end", default=None): vol.Any(TIME_SCHEMA, None),
        vol.Optional("end_plus_days", default=0):
            vol.All(int, vol.Range(min=0)),
        vol.Optional("years"): RANGE_STRING_SCHEMA,
        vol.Optional("months"): RANGE_STRING_SCHEMA,
        vol.Optional("days"): RANGE_STRING_SCHEMA,
        vol.Optional("weeks"): RANGE_STRING_SCHEMA,
        vol.Optional("weekdays"): RANGE_STRING_SCHEMA,
        vol.Optional("start_date"): PARTIAL_DATE_SCHEMA,
        vol.Optional("end_date"): PARTIAL_DATE_SCHEMA,
    },
))
SCHEDULE_SCHEMA = vol.Schema(vol.All(
    lambda v: v or [],
    [SCHEDULE_RULE_SCHEMA],
))

SCHEDULE_SNIPPETS_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {vol.Extra: SCHEDULE_SCHEMA},
))

ROOM_SCHEMA = vol.Schema(vol.All(
    lambda v: v or {},
    {
        "friendly_name": str,
        vol.Optional("replicate_changes", default=True): bool,
        vol.Optional("reschedule_delay", default=0):
            vol.All(int, vol.Range(min=0)),
        vol.Optional("thermostats", default=dict): DICTS_IN_DICT_SCHEMA,
        vol.Optional("window_sensors", default=dict): DICTS_IN_DICT_SCHEMA,
        vol.Optional("schedule", default=list):
            SCHEDULE_SCHEMA,
    },
))

CONFIG_SCHEMA = vol.Schema(vol.All(
    vol.Schema({
        vol.Optional("heaty_id", default="default"): str,
        vol.Optional("untrusted_temp_expressions", default=False): bool,
        vol.Optional("master_switch", default=None):
            vol.Any(ENTITY_ID_SCHEMA, None),
        vol.Optional("off_temp", default=expr.OFF): TEMP_SCHEMA,
        vol.Optional("temp_expression_modules", default=dict):
            TEMP_EXPRESSION_MODULES_SCHEMA,
        vol.Optional("thermostat_defaults", default=dict):
            lambda v: v or {},
        vol.Optional("window_sensor_defaults", default=dict):
            lambda v: v or {},
        vol.Optional("schedule_prepend", default=list):
            SCHEDULE_SCHEMA,
        vol.Optional("schedule_append", default=list):
            SCHEDULE_SCHEMA,
        vol.Optional("schedule_snippets", default=dict):
            SCHEDULE_SNIPPETS_SCHEMA,
        vol.Optional("rooms", default=dict): vol.All(
            lambda v: v or {},
            {vol.Extra: ROOM_SCHEMA},
        ),
    }, extra=True),
    config_post_hook,
))



"""
This module implements the Schedule and Rule classes.
"""

import typing as T  # pylint: disable=unused-import

import datetime

from . import expr, util


class Rule:
    """A rule that can be added to a schedule."""

    # names of schedule rule constraints to be fetched from a rule definition
    CONSTRAINTS = ("years", "months", "days", "weeks", "weekdays",
                   "start_date", "end_date")

    def __init__(
            self, temp_expr: expr.EXPR_TYPE,
            start_time: datetime.time = None, end_time: datetime.time = None,
            end_plus_days: int = 0, constraints: T.Dict[str, T.Any] = None
        ) -> None:
        if start_time is None:
            # make it midnight
            start_time = datetime.time(0, 0)
        self.start_time = start_time

        if end_time is None:
            # make it midnight (00:00 of the next day)
            end_time = datetime.time(0, 0)
            end_plus_days += 1
        self.end_time = end_time

        if end_time < start_time and end_plus_days == 0:
            end_plus_days = 1
        self.end_plus_days = end_plus_days

        if constraints is None:
            constraints = {}
        self.constraints = constraints

        if isinstance(temp_expr, str):
            temp_expr = temp_expr.strip()
        self.temp_expr_raw = temp_expr
        try:
            temp = expr.Temp(temp_expr)
        except ValueError:
            # this is a temperature expression, precompile it
            self.temp_expr = compile(temp_expr, "temp_expr", "eval")  # type: expr.EXPR_TYPE
        else:
            self.temp_expr = temp

    def check_constraints(self, date: datetime.date) -> bool:
        """Checks all constraints of this rule against the given date."""

        # pylint: disable=too-many-return-statements

        year, week, weekday = date.isocalendar()
        for constraint, allowed in self.constraints.items():
            if constraint == "years" and year not in allowed:
                return False
            if constraint == "months" and date.month not in allowed:
                return False
            if constraint == "days" and date.day not in allowed:
                return False
            if constraint == "weeks" and week not in allowed:
                return False
            if constraint == "weekdays" and weekday not in allowed:
                return False
            if constraint == "start_date" and \
               date < util.build_date_from_constraint(allowed, date, 1):
                return False
            if constraint == "end_date" and \
               date > util.build_date_from_constraint(allowed, date, -1):
                return False
        return True


class Schedule:
    """Holds the schedule for a room with all its rules."""

    def __init__(self) -> None:
        self.items = []  # type: T.List[T.Union[Rule, Schedule]]

    def unfold(self) -> T.Iterator[Rule]:
        """Returns an iterator over all rules of this schedule. Included
        sub-schedules are replaced by the rules they contain."""

        for item in self.items:
            if isinstance(item, Rule):
                yield item
            elif isinstance(item, Schedule):
                for rule in item.unfold():
                    yield rule

    def matching_rules(self, when: datetime.datetime) -> T.Iterator[Rule]:
        """Returns an iterator over all rules of the schedule that are
        valid at the time represented by the given datetime object,
        keeping the order from the items list. Rules of sub-schedules
        are included."""

        _time = when.time()
        for rule in self.unfold():
            days_back = -1
            found_start_day = False
            while days_back < rule.end_plus_days:
                days_back += 1
                # starts with days=0 (meaning the current date)
                _date = when.date() - datetime.timedelta(days=days_back)

                found_start_day = found_start_day or \
                                  rule.check_constraints(_date)
                if not found_start_day:
                    # try next day
                    continue

                # in first loop run, rule has to start today and not
                # later than now (rule start <= when.time())
                if days_back == 0 and rule.start_time > _time:
                    # maybe there is a next day to try out
                    continue

                # in last loop run, rule is going to end today and that
                # has to be later than now (rule end > when.time())
                if days_back == rule.end_plus_days and rule.end_time <= _time:
                    # rule finally disqualified
                    break

                # rule matches!
                yield rule
                break

    def next_schedule_datetime(
            self, now: datetime.datetime
    ) -> T.Optional[datetime.datetime]:
        """Returns a datetime object with the time at which the next
        re-scheduling should be done. now should be a datetime object
        containing the current date and time.
        None is returned in case there are no rules in the schedule."""

        times = set()
        for rule in self.unfold():
            times.add(rule.start_time)
            times.add(rule.end_time)

        if not times:
            # no rules in schedule
            return None

        current_time = now.time()
        today = now.date()
        tomorrow = today + datetime.timedelta(days=1)

        def map_func(_time: datetime.time) -> datetime.datetime:
            """Maps a time object to a datetime containing the next
            occurrence of that time. Midnight transitions are handled
            correctly."""

            if _time < current_time:
                # midnight transition
                return datetime.datetime.combine(tomorrow, _time)

            return datetime.datetime.combine(today, _time)

        return min(map(map_func, times))



#!/usr/bin/env python3


import os
from setuptools import find_packages, setup

from hass_apps import __version__


def read_file(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name = "hass-apps",
    version = __version__,
    description = "A collection of useful apps for AppDaemon + "
                  "Home Assistant.",
    long_description = read_file("README.rst"),
    url = "https://github.com/efficiosoft/hass-apps",
    author = "Robert Schindler",
    author_email = "r.schindler@efficiosoft.com",
    license = "Apache 2.0",
    packages = find_packages("."),
    package_data = {
        "hass_apps": ["data/*"],
    },
    install_requires = [
        "appdaemon >= 2.1.12",
        "voluptuous >= 0.11.1",
    ],
    zip_safe = False,
)



"""
A simple app for connecting motion sensors to arbitrary entities
in Home Assistant, so that motion causes these entities to be turned
on or off.
"""

import typing as T

import voluptuous as vol

from .. import common
from . import __version__


CONSTRAINTS_SCHEMA = vol.Schema({
    vol.Extra: str,
})

CONTROL_SCHEMA = vol.Schema({
    vol.Optional("invert", default=False): bool,
})
CONTROLS_SCHEMA = vol.Schema({
    vol.Extra: lambda v: CONTROL_SCHEMA(v or {}),
})

SENSOR_SCHEMA = vol.Schema({
    vol.Optional("global_constraints", default=True): bool,
    vol.Optional("constraints", default=dict): lambda v: CONSTRAINTS_SCHEMA(v or {}),
    vol.Optional("on_state", default="on"): str,
    vol.Optional("on_delay", default=0): int,
    vol.Optional("off_delay", default=0): int,
    vol.Optional("controls", default=dict): lambda v: CONTROLS_SCHEMA(v or {}),
})
SENSORS_SCHEMA = vol.Schema({
    vol.Extra: lambda v: SENSOR_SCHEMA(v or {}),
})

CONFIG_SCHEMA = vol.Schema({
    vol.Optional("debug", default=False): bool,
    vol.Optional("constraints", default=dict): lambda v: CONSTRAINTS_SCHEMA(v or {}),
    vol.Optional("sensors", default=dict): lambda v: SENSORS_SCHEMA(v or {}),
}, extra=True)


class MotionLightApp(common.App):
    """
    A simple app for connecting motion sensors to arbitrary entities
    in Home Assistant, so that motion causes these entities to be turned
    on or off.
    """

    class Meta(common.App.Meta):
        # pylint: disable=missing-docstring
        name = "motion_light"
        version = __version__
        config_schema = CONFIG_SCHEMA

    def initialize_inner(self) -> None:
        """Parses the configuration and sets up state listeners."""

        # pylint: disable=attribute-defined-outside-init

        for sensor, sensor_data in self.cfg["sensors"].items():
            sensor_data["turned_on"] = False
            if sensor_data["global_constraints"]:
                for key, value in self.cfg["constraints"].items():
                    sensor_data["constraints"].setdefault(key, value)

        for sensor, sensor_data in self.cfg["sensors"].items():
            self.listen_state(self._sensor_state_cb, sensor,
                              new=sensor_data["on_state"],
                              duration=sensor_data["on_delay"],
                              **sensor_data["constraints"])
            self.listen_state(self._sensor_state_cb, sensor,
                              old=sensor_data["on_state"],
                              duration=sensor_data["off_delay"])

    def _sensor_state_cb(
            self, sensor: str, attr: str, old: T.Any, new: T.Any, kwargs: dict
    ) -> None:
        """Is called whenever a motion sensor changes between on and off."""

        sensor_data = self.cfg["sensors"][sensor]
        is_on = new == sensor_data["on_state"]

        self.log("[{}] Motion {}ed"
                 .format(sensor, "start" if is_on else "end"),
                 level="DEBUG", prefix=common.LOG_PREFIX_INCOMING)

        if not is_on and not sensor_data["turned_on"]:
            # motion ended, but nothing has been turned on before
            # (e.g. due to constraints), hence nothing should be
            # turned off either
            self.log("[{}] Ignoring motion end because not turned on before."
                     .format(sensor),
                     level="DEBUG")
            return

        sensor_data["turned_on"] = is_on

        for entity, entity_data in sensor_data["controls"].items():
            turn_on = is_on ^ entity_data["invert"]
            self.log("Turning {} {}"
                     .format(entity, "on" if turn_on else "off"),
                     level="DEBUG", prefix=common.LOG_PREFIX_OUTGOING)
            domain = self.split_entity(entity)[0]
            service = "{}/turn_{}".format(domain, "on" if turn_on else "off")
            self.call_service(service, entity_id=entity)



"""
A highly-configurable, comfortable to use Home Assistant / appdaemon app
that controls thermostats based on a schedule while still facilitating
manual intervention at any time.
"""

import typing as T
import types  # pylint: disable=unused-import

import datetime
import importlib

from .. import common
from . import __version__, config, expr, util
from . import room as _room


__all__ = ["HeatyApp"]


class HeatyApp(common.App):
    """The Heaty app class for AppDaemon."""

    class Meta(common.App.Meta):
        # pylint: disable=missing-docstring
        name = "heaty"
        version = __version__
        config_schema = config.CONFIG_SCHEMA

    def __init__(self, *args: T.Any, **kwargs: T.Any) -> None:
        self.app = self
        self.cfg = None
        self.rooms = []  # type: T.List[_room.Room]
        self.publish_state_timer = None
        self.temp_expression_modules = {}  # type: T.Dict[str, types.ModuleType]
        super().__init__(*args, **kwargs)

    @util.modifies_state
    def initialize_inner(self) -> None:
        """Checks the configuration, initializes all timers, state and
        event callbacks and sets temperatures in all rooms according
        to the configured schedules."""

        heaty_id = self.cfg["heaty_id"]
        self.log("Heaty id is: {}".format(repr(heaty_id)))
        heaty_id_kwargs = {}
        if heaty_id != "default":
            heaty_id_kwargs["heaty_id"] = heaty_id

        self.log("Importing modules for use in temperature expressions.",
                 level="DEBUG")
        for mod_name, mod_data in self.cfg["temp_expression_modules"].items():
            as_name = util.escape_var_name(mod_data.get("as", mod_name))
            self.log("Importing module {} as {}."
                     .format(repr(mod_name), repr(as_name)),
                     level="DEBUG")
            try:
                mod = importlib.import_module(mod_name)
            except Exception as err:  # pylint: disable=broad-except
                self.log("Error while importing module {}: {}"
                         .format(repr(mod_name), repr(err)),
                         level="ERROR")
                self.log("Module won't be available.", level="ERROR")
            else:
                self.temp_expression_modules[as_name] = mod

        for room in self.rooms:
            room.initialize()

        self.log("Registering event listener for heaty_reschedule.",
                 level="DEBUG")
        self.listen_event(self._reschedule_event_cb, "heaty_reschedule",
                          **heaty_id_kwargs)

        self.log("Registering event listener for heaty_set_temp.",
                 level="DEBUG")
        self.listen_event(self._set_temp_event_cb, "heaty_set_temp",
                          **heaty_id_kwargs)

        master_switch = self.cfg["master_switch"]
        if master_switch:
            self.log("Registering state listener for master switch {}."
                     .format(master_switch),
                     level="DEBUG")
            self.listen_state(self._master_switch_cb, master_switch)

        if self.master_switch_enabled():
            for room in self.rooms:
                if not room.check_for_open_window():
                    room.set_scheduled_temp()
        else:
            self.log("Master switch is off, not setting temperatures "
                     "initially.")

    @util.modifies_state
    def _master_switch_cb(
            self, entity: str, attr: str, old: T.Any, new: T.Any, kwargs: dict
    ) -> None:
        """Is called when the master switch is toggled.
        If turned on, it sets the scheduled temperatures in all rooms.
        If switch is turned off, all re-schedule timers are cancelled
        and temperature is set to self.cfg["off_temp"] everywhere."""

        self.log("Master switch turned {}.".format(new),
                 prefix=common.LOG_PREFIX_INCOMING)
        for room in self.rooms:
            if new == "on":
                room.set_scheduled_temp()
            else:
                room.cancel_reschedule_timer()
                room.set_temp(self.cfg["off_temp"], scheduled=False)
                # invalidate cached temp/rule
                room.current_schedule_temp = None
                room.current_schedule_rule = None

    def _publish_state_timer_cb(self, kwargs: dict) -> None:
        """Runs when a publish_state timer fires."""

        self.publish_state_timer = None
        self.publish_state()

    def _reschedule_event_cb(
            self, event: str, data: dict, kwargs: dict
    ) -> None:
        """This callback executes when a heaty_reschedule event is received.
        data may contain a "room_name", which limits the re-scheduling
        to the given room."""

        if not self.master_switch_enabled():
            self.log("Ignoring re-schedule event because master "
                     "switch is off.")
            return

        room_name = data.get("room_name")
        if room_name:
            room = self.get_room(room_name)
            if not room:
                self.log("Ignoring heaty_reschedule event for "
                         "unknown room {}.".format(room_name),
                         level="WARNING")
                return
            rooms = [room]
        else:
            rooms = self.rooms

        self.log("Re-schedule event received for rooms: {}"
                 .format(", ".join([str(room) for room in rooms])),
                 prefix=common.LOG_PREFIX_INCOMING)

        for room in rooms:
            # delay for 6 seconds to avoid re-scheduling multiple
            # times if multiple events come in shortly
            room.update_reschedule_timer(reschedule_delay=0.1,
                                         force=True)

    def _set_temp_event_cb(
            self, event: str, data: dict, kwargs: dict
    ) -> None:
        """This callback executes when a heaty_set_temp event is received.
        data must contain a "room_name" and a "temp", which may also
        be a temperature expression. "force_resend" is optional and
        False by default. If it is set to True, the temperature is
        re-sent to the thermostats even if it hasn't changed due to
        Heaty's records."""

        try:
            room_name = data["room_name"]
            temp_expr = data["temp"]
            reschedule_delay = data.get("reschedule_delay")
            if not isinstance(reschedule_delay, (type(None), float, int)):
                raise TypeError()
            if isinstance(reschedule_delay, (float, int)) and \
               reschedule_delay < 0:
                raise ValueError()
        except (KeyError, TypeError, ValueError):
            self.log("Ignoring heaty_set_temp event with invalid data: {}"
                     .format(repr(data)),
                     level="WARNING")
            return

        room = self.get_room(room_name)
        if not room:
            self.log("Ignoring heaty_set_temp event for unknown "
                     "room {}.".format(room_name),
                     level="WARNING")
            return

        if not self.cfg["untrusted_temp_expressions"] and \
           expr.Temp.parse_temp(temp_expr) is None:
            self.log("Ignoring heaty_set_temp event with an "
                     "untrusted temperature expression. "
                     "(untrusted_temp_expressions = false)",
                     level="WARNING")
            return

        room.notify_set_temp_event(
            temp_expr, force_resend=bool(data.get("force_resend")),
            reschedule_delay=reschedule_delay
        )

    def get_room(self, room_name: str) -> T.Optional[_room.Room]:
        """Returns the room with given name or None, if no such room
        exists."""

        for room in self.rooms:
            if room.name == room_name:
                return room
        return None

    def master_switch_enabled(self) -> bool:
        """Returns the state of the master switch or True if no master
        switch is configured."""
        master_switch = self.cfg["master_switch"]
        if master_switch:
            return self.get_state(master_switch) == "on"  # type: ignore
        return True

    def publish_state(self) -> None:
        """Publishes Heaty's current state to AppDaemon for use by
        dashboards."""

        # shortcuts to make expr.Temp and datetime.time objects json
        # serializable
        unpack_temp = lambda t: t.value if isinstance(t, expr.Temp) else t  # type: T.Callable[[T.Any], T.Any]  # pylint: disable=line-too-long
        unpack_time = lambda t: t.strftime(util.TIME_FORMAT) \
                if isinstance(t, datetime.time) else None  # type: T.Callable[[T.Any], T.Any]

        self.log("Publishing state to AppDaemon.",
                 level="DEBUG", prefix=common.LOG_PREFIX_OUTGOING)

        attrs = {
            "heaty_id": self.cfg["heaty_id"],
            "master_switch_enabled": self.master_switch_enabled(),
        }

        rooms = {}
        for room in self.rooms:
            schedule = room.schedule
            now = self.datetime()
            if schedule:
                next_schedule_datetime = schedule.next_schedule_datetime(now)
            else:
                next_schedule_datetime = None
            if next_schedule_datetime:
                # pylint: disable=line-too-long
                next_schedule_time = next_schedule_datetime.time()  # type: T.Optional[datetime.time]
            else:
                next_schedule_time = None

            rooms[room.name] = {
                "friendly_name": room.cfg.get("friendly_name", room.name),
                "wanted_temp": unpack_temp(room.wanted_temp),
                "current_schedule_temp":
                    unpack_temp(room.current_schedule_temp),
                "next_schedule_time": unpack_time(next_schedule_time),
            }

            thermostats = {}
            for therm in room.thermostats:
                thermostats[therm.entity_id] = {
                    "current_temp": unpack_temp(therm.current_temp),
                    "resend_timer": bool(therm.resend_timer),
                }
            rooms[room.name]["thermostats"] = thermostats

            window_sensors = {}
            for sensor in room.window_sensors:
                window_sensors[sensor.entity_id] = {
                    "open": sensor.is_open(),
                }
            rooms[room.name]["window_sensors"] = window_sensors
        attrs["rooms"] = rooms

        entity_id = "appdaemon.heaty_{}".format(self.cfg["heaty_id"])
        state = {"state": None, "attributes": attrs}
        self.set_app_state(entity_id, state)

    def update_publish_state_timer(self) -> None:
        """Sets the publish_state timer to fire in 1 second, if not
        running already."""

        self.log("Called update_publish_state_timer.", level="DEBUG")

        if not self.publish_state_timer:
            timer = self.run_in(self._publish_state_timer_cb, 1)
            self.publish_state_timer = timer



"""
Utility functions that are used everywhere inside Heaty.
"""

import typing as T

import datetime
import re
import time


# matches any character that is not allowed in Python variable names
INVALID_VAR_NAME_CHAR_PATTERN = re.compile(r"[^0-9A-Za-z_]")
# regexp pattern matching a range like 3-7 without spaces
RANGE_PATTERN = re.compile(r"^(\d+)\-(\d+)$")
# strftime-compatible format string for military time
TIME_FORMAT = "%H:%M"


def escape_var_name(name: str) -> str:
    """Converts the given string to a valid Python variable name.
    All unsupported characters are replaced by "_". If name would
    start with a digit, "_" is put infront."""

    name = INVALID_VAR_NAME_CHAR_PATTERN.sub("_", name)
    digits = tuple([str(i) for i in range(10)])
    if name.startswith(digits):
        name = "_" + name
    return name

def expand_range_string(range_string: T.Union[float, int, str]) -> T.Set[int]:
    """Expands strings of the form '1,2-4,9,11-12 to set(1,2,3,4,9,11,12).
    Any whitespace is ignored. If a float or int is given instead of a
    string, a set containing only that, converted to int, is returned."""

    if isinstance(range_string, (float, int)):
        return set([int(range_string)])

    numbers = set()
    for part in "".join(range_string.split()).split(","):
        match = RANGE_PATTERN.match(part)
        if match is not None:
            for i in range(int(match.group(1)), int(match.group(2)) + 1):
                numbers.add(i)
        else:
            numbers.add(int(part))
    return numbers

def build_date_from_constraint(
        constraint: T.Dict[str, int], default_date: datetime.date,
        direction: int = 0
) -> datetime.date:
    """Builds and returns a datetime.date object from the given constraint,
    taking missing values from the given default_date.
    In case the date is not valid (e.g. 2017-02-29), a ValueError is
    raised, unless a number has been given for direction, in which case
    the next/previous valid date will be chosen, depending on the sign
    of direction."""

    fields = {}
    for field in ("year", "month", "day"):
        fields[field] = constraint.get(field, getattr(default_date, field))

    while True:
        try:
            return datetime.date(**fields)
        except ValueError:
            if direction > 0:
                fields["day"] += 1
            elif direction < 0:
                fields["day"] -= 1
            else:
                raise

            # handle month/year transitions correctly
            if fields["day"] < 1:
                fields["day"] = 31
                fields["month"] -= 1
            elif fields["day"] > 31:
                fields["day"] = 1
                fields["month"] += 1
            if fields["month"] < 1:
                fields["month"] = 12
                fields["year"] -= 1
            elif fields["month"] > 12:
                fields["month"] = 1
                fields["year"] += 1

def format_time(when: datetime.time, format_str: str = TIME_FORMAT) -> str:
    """Returns a string representing the given datetime.time object.
    If no strftime-compatible format is provided, the default is used."""

    return when.strftime(format_str)

def modifies_state(func: T.Callable) -> T.Callable:
    """This decorator calls update_publish_state_timer() after the
    method decorated with it ran. It may only be used for non-static
    methods of a class containing the app as "app" attribute, because it
    fetches the "app" attribute from the method's first argument."""

    def _new_func(self: T.Any, *args: T.Any, **kwargs: T.Any) -> T.Any:
        result = func(self, *args, **kwargs)
        self.app.update_publish_state_timer()
        return result

    _new_func.__name__ = func.__name__
    _new_func.__doc__ = func.__doc__
    _new_func.__annotations__ = func.__annotations__
    _new_func.__dict__.update(func.__dict__)
    return _new_func

def parse_time_string(
        time_str: str, format_str: str = TIME_FORMAT
) -> datetime.time:
    """Parses a string of the given strptime-compatible format into
    a datetime.time object. If the string has an invalid format, a
    ValueError is raised. If no format is provided, the default will
    be used."""

    # remove whitespace
    time_str = "".join(time_str.split())
    t_struct = time.strptime(time_str, format_str)
    return datetime.time(t_struct.tm_hour, t_struct.tm_min, t_struct.tm_sec)



"""
Common helpers and functionality used by all apps.
"""

import typing as T

import copy

try:
    from appdaemon.plugins.hass import hassapi
except ImportError:
    from appdaemon import appapi
    AppBase = appapi.AppDaemon
    _IS_AD3 = False
else:
    AppBase = hassapi.Hass
    _IS_AD3 = True


# types of log messages, used for determining the prefix
LOG_PREFIX_NONE = ""
LOG_PREFIX_STATUS = "---"
LOG_PREFIX_WARNING = "!!!"
LOG_PREFIX_INCOMING = "-->"
LOG_PREFIX_OUTGOING = "<--"


class App(AppBase):
    """
    This is a sub-class of hassapi.Hass (for appdaemon >= 3.0.0) or
    appapi.AppDaemon (for appdaemon < 3.0.0) which adds some common
    functionality. It's used by all apps included in hass_apps.
    """

    # will be True for appdaemon >= 3, False otherwise
    _is_ad3 = _IS_AD3

    class Meta:
        """
        A class holding information about the app such as name and version.
        This information is used for logging, for instance.
        """

        name = "UNCONFIGURED"
        version = "0.0.0"
        config_schema = None  # type: T.Optional[T.Callable]

    def log(  # pylint: disable=arguments-differ
            self, msg: str, level: str = "INFO",
            prefix: T.Optional[str] = None
    ) -> None:
        """Wrapper around super().log() which changes the log level
        from DEBUG to INFO if debug config option is enabled.
        It also adds an appropriate prefix to the log message."""

        level = level.upper()
        if level == "DEBUG" and self.args and self.args.get("debug"):
            level = "INFO"

        if prefix is None:
            if level in ("DEBUG", "INFO"):
                prefix = LOG_PREFIX_STATUS
            elif level in ("WARNING", "ERROR"):
                prefix = LOG_PREFIX_WARNING

        if prefix:
            msg = "{} {}".format(prefix, msg)

        super().log(msg, level=level)

    def initialize(self) -> None:
        """Parses the configuration and logs that initialization
        started/finished. The real work should be done in
        initialize_inner()."""

        # pylint: disable=attribute-defined-outside-init

        self.log("{} v{} initialization started"
                 .format(self.Meta.name, self.Meta.version))

        if callable(self.Meta.config_schema):
            self.log("Validating the app's configuration.", level="DEBUG")
            cfg = copy.deepcopy(self.args)
            # Make the app object available during config validation.
            cfg["_app"] = self
            self.cfg = self.Meta.config_schema(cfg)  # pylint: disable=not-callable

        self.initialize_inner()

        self.log("Initialization done")

    def initialize_inner(self) -> None:
        """Overwrite this stub to do the real initialization of the
        particular app."""

        pass

    def set_app_state(
            self, entity_id: str, state: T.Dict[str, T.Any]
    ) -> None:
        """A wrapper to make the new appdaemon.AppDaemon.set_app_state
        available under the appdaemon 2 interface."""

        if self._is_ad3:
            self.AD.set_app_state(entity_id, state)
        else:
            super().set_app_state(entity_id, state)  # pylint: disable=no-member
